# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021 The Amoni project team members listed at
# https://github.com/anvilistas/amoni/graphs/contributors
#
# This software is published at https://github.com/anvilistas/amoni
import os
from pathlib import Path
from typing import Dict, List

import keyring
import pygit2
from cookiecutter.main import cookiecutter
from python_on_whales import docker
from yaml import dump, load

try:
    from yaml import CDumper as Dumper
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Dumper, Loader

__version__ = "0.0.9"

COOKIECUTTER_URL = "https://github.com/anvilistas/amoni-cookiecutter.git"
ANVIL_CONFIG_FILE = Path("app", "config.yaml")
TABLE_STUB_FILE = Path("anvil-stubs", "tables", "app_tables.pyi")


def _commit_all(
    message, repo=None, ref=None, author=None, committer=None, tree=None, parents=None
):
    cwd = os.getcwd()
    os.chdir(repo.path)
    repo = repo or pygit2.Repository(".")
    repo.index.add_all()
    repo.index.write()

    ref = ref or repo.head.name
    author = author or repo.default_signature
    committer = committer or repo.default_signature
    tree = tree or repo.index.write_tree()
    parents = parents if parents is not None else [repo.head.target]
    repo.create_commit(ref, author, committer, message, tree, parents)

    os.chdir(cwd)


def _get_app_config(app: str) -> Dict:
    config_file = Path("app", app, "anvil.yaml")
    return load(config_file.open(), Loader=Loader)


def _save_app_config(app: str, config: Dict) -> None:
    config_file = Path("app", app, "anvil.yaml")
    dump(config, config_file.open("w"), Dumper=Dumper)


def init(directory: Path, app: str) -> None:
    """Initialise an amoni project

    Parameters
    ----------
    directory
        The full path of the amoni project folder to create
    app
        The name of folder within the 'app' folder which contains the app to be run
    """
    cookiecutter(
        COOKIECUTTER_URL,
        no_input=True,
        output_dir=directory.parent,
        extra_context={"project_name": directory.name, "app_folder_name": app},
    )
    repo = pygit2.init_repository(directory)
    _commit_all("Initial commit", repo=repo, ref="HEAD", parents=[])


def pull_image(name: str) -> None:
    """Pull a docker image from the github registry

    Parameters
    ----------
    name
        The name of the image to pull
    """
    docker.compose.pull([name])


def build_image(name: str) -> None:
    """Build a docker image
    Parameters
    ----------
    name
        The name of the image to pull
    """
    docker.compose.build([name], pull=True)


def start_service(name: str, detach: bool) -> None:
    """Start a given service

    Parameters
    ----------
    name
        The name of the service to start
    detach
        Whether to detach from the service console
    """
    docker.compose.up([name], detach=detach)


def stop_services() -> None:
    """Stop all amoni services"""
    docker.compose.down()


def run_service(name: str) -> None:
    """Run a given service

    Parameters
    ----------
    name
        The name of the service to start
    """
    docker.compose.run(name)


class AmoniRemoteCallbacks(pygit2.RemoteCallbacks):
    def credentials(self, url, user, allowed_types):
        if allowed_types & pygit2.credentials.GIT_CREDENTIAL_SSH_KEY:
            credentials = {
                key: keyring.get_password(url, key)
                for key in ("username", "pubkey", "privkey", "passphrase")
            }
            if user is not None:
                credentials["username"] = user
            return pygit2.Keypair(**credentials)
        return None


def add_submodule(url: str, path: Path, name: str) -> None:
    """Add a submodule to the current repository

    Parameters
    ----------
    url
        The url of the submodule to add
    path
        The directory where the submodule should be added
    name
        The name of the submodule
    """
    repo = pygit2.Repository(".")
    repo.add_submodule(url, path, callbacks=AmoniRemoteCallbacks())
    _commit_all(f"Add {name} submodule", repo=repo)


def set_app(name: str) -> None:
    """Set the app to be run by the anvil app server

    Parameters
    ----------
    name
        The name of the app
    """
    anvil_config = load(ANVIL_CONFIG_FILE.open(), Loader=Loader)
    anvil_config["app"] = Path("/", "app", name).as_posix()
    dump(anvil_config, ANVIL_CONFIG_FILE.open("w"), Dumper=Dumper)
    _commit_all(f"Set {name} as the anvil app")


def set_dependency(id: str, name: str) -> None:
    """Set an app to be a dependency

    Parameters
    ----------
    id
        The id of the dependency app
    name
        The name of the dependency app
    """
    anvil_config = load(ANVIL_CONFIG_FILE.open(), Loader=Loader)
    try:
        anvil_config["dep-id"][id] = name
    except KeyError:
        anvil_config["dep-id"] = {id: name}
    dump(anvil_config, ANVIL_CONFIG_FILE.open("w"), Dumper=Dumper)
    _commit_all(f"Set {name} as a dependency")


def generate_table_stubs(app: str, target: Path = TABLE_STUB_FILE) -> None:
    """Generate stub entries for app tables in anvil.yaml

    Parameters
    ----------
    app
        The name of the app
    target
        The stub file where the entries should be added
    """
    config = _get_app_config(app)
    tables = config["db_schema"]
    table_definition = [f"{t}: Table\n" for t in tables]
    content = ["from anvil.tables import Table\n\n"] + table_definition
    with Path(target).open("w") as f:
        f.truncate(0)
        f.writelines(content)


def build_theme(app: str) -> None:
    """Build the theme for the given app

    Parameters
    ----------
    app
        The name of the app
    """
    docker.compose.run("theme_builder", [app])


def add_table(
    app: str,
    name: str,
    client_permissions: str = "none",
    server_permissions: str = "full",
    columns: List = None,
):
    columns = columns if columns is None else []
    config = _get_app_config(app)
    table = {
        "title": name,
        "client": client_permissions,
        "server": server_permissions,
        "columns": columns,
    }
    try:
        config["db_schema"][name] = table
    except KeyError:
        config["db_schema"] = {name: table}
    _save_app_config(app, config)
    _commit_all(f"Add {name} data table")


def add_column(app: str, table: str, name: str, data_type: str, target: str = None):
    config = _get_app_config(app)
    column = {"name": name, "admin_ui": {"width": 200}, "type": data_type}
    if not target:
        column["target"] = target
    config["db_schema"][table]["columns"].append(column)
    _save_app_config(app, config)
    _commit_all(f"Add {name} column to {table} data table")



# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021 The Amoni project team members listed at
# https://github.com/anvilistas/amoni/graphs/contributors
#
# This software is published at https://github.com/anvilistas/amoni
from pathlib import Path

import typer
from cookiecutter.exceptions import OutputDirExistsException

from .. import api
from . import app, echo, theme

__version__ = "0.0.9"

cmd = typer.Typer()
cmd.add_typer(app.cmd, name="app", help="Manage anvils apps and dependencies")
cmd.add_typer(theme.cmd, name="theme", help="Build the theme.css for an app")


@cmd.callback()
def main():
    pass


@cmd.command()
def init(
    directory: Path = typer.Argument(
        ..., file_okay=False, resolve_path=True, help="Directory to initialiase"
    ),
    app: str = typer.Argument("hello_world", help="App Folder Name"),
):
    """Create the amoni folder structure and initialise a git repo there"""
    try:
        api.init(directory, app)
        echo.progress(f"Amoni project created in {directory}")
        echo.done()
    except OutputDirExistsException:
        echo.error(f"Error creating project. {directory} already exists")


@cmd.command()
def start(
    update: bool = typer.Option(False, help="Whether to update the docker images"),
    launch: bool = typer.Option(True, help="Whether to launch the app"),
):
    """Start the anvil app and db servers"""
    if update:
        typer.echo("Rebuilding server images")
        api.build_image("app")
        api.pull_image("db")
    with echo.working("Starting anvil app and database servers"):
        api.start_service("app", detach=True)
    echo.progress("Your app is available at http://localhost:3030")
    if launch:
        typer.launch("http://localhost:3030")
    echo.done()


@cmd.command()
def stop():
    """Stop the anvil app and db servers"""
    with echo.working("Stopping anvil app and database servers"):
        api.stop_services()
    echo.done()


@cmd.command()
def test():
    """Run the test suite"""
    echo.progress("Checking for newer images")
    service = "test_runner"
    api.build_image(service)
    api.run_service(service)


@cmd.command()
def stubs(app: str = typer.Argument(..., help="App folder name")):
    """Generate stubs for the database"""
    api.generate_table_stubs(app)
    echo.progress(f"Created table definitions in {api.TABLE_STUB_FILE}")
    echo.done()



# SPDX-License-Identifier: MIT
#
# Copyright (c) 2021 The Amoni project team members listed at
# https://github.com/anvilistas/amoni/graphs/contributors
#
# This software is published at https://github.com/anvilistas/amoni
import time
from contextlib import contextmanager
from functools import partial
from threading import Event, Thread

import typer

__version__ = "0.0.9"

CLEAR_LINE = "\r\033[K"
EMOJI_ANIMATION = {"frames": ("⣷", "⣯", "⣟", "⡿", "⢿", "⣻", "⣽", "⣾"), "interval": 0.1}
ASCII_ANIMATION = {"frames": ("   ", ".  ", ".. ", "..."), "interval": 0.5}
DECORATIONS = {"error": "⛔", "done": "✨️"}


def _emojis_work():
    emojis = "".join(EMOJI_ANIMATION) + "".join(DECORATIONS.values())
    try:
        emojis.encode()
        return True
    except UnicodeEncodeError:
        return False


def _decorate(message, decoration):
    return f"{DECORATIONS[decoration]} {message}" if _emojis_work() else message


def error(message):
    typer.secho(_decorate(message, "error"), fg=typer.colors.RED, err=True)


def done():
    typer.secho(_decorate("Done!", "done"), fg=typer.colors.GREEN)


progress = partial(typer.secho, fg=typer.colors.CYAN)


def _animate(message, event):
    animation = EMOJI_ANIMATION if _emojis_work() else ASCII_ANIMATION
    while not event.is_set():
        for frame in animation["frames"]:
            progress(f"{frame} {message}", nl=False)
            time.sleep(animation["interval"])
            progress(CLEAR_LINE, nl=False)


@contextmanager
def working(message):
    """Prints a working message"""
    event = Event()
    kwargs = {"message": message, "event": event}
    thread = Thread(target=_animate, kwargs=kwargs)
    thread.start()

    try:
        yield
    finally:
        event.set()
        thread.join()
